@@ -974,7 +974,7 @@ SEXP lmer_invert(SEXP x)
 	    error(_("Leading minor of size %d of downdated X'X,is indefinite"),
 		  i + 1);
 
-	/* RZX := - RZX %*% RXX */
+				/* RZX := - RZX %*% RXX */
 	F77_CALL(dtrmm)("R", "U", "N", "N", &dims[0], &ncX, &minus1,
 			RXX, &ncX, RZX, &dims[0]);
 	for(i = 0; i < nf; i++) {
@@ -984,7 +984,7 @@ SEXP lmer_invert(SEXP x)
 		*RZXi = RZX + Gp[i];
 
 	    nlevs[i] = (Gp[i+1] - Gp[i])/nci;
-	    /* D_i := D_i^{-1}; RZX_i := D_i %*% RZX_i */
+				/* D_i := D_i^{-1}; RZX_i := D_i %*% RZX_i */
 	    if (nci == 1) {
 		for (j = 0; j < nlevs[i]; j++) {
 		    Di[j] = 1./Di[j];
@@ -1002,7 +1002,7 @@ SEXP lmer_invert(SEXP x)
 	    }
 	}
 
-	/* RZX := L^{-T} %*% RZX */
+				/* RZX := L^{-T} %*% RZX */
 	internal_sm(LFT, TRN, nf, Gp, ncX, 1.0, LP, RZX, dims[0]);
 
 	alloc_tmp_ind(nf, nc, nlevs, ParP, tmp, ind);
@@ -1490,6 +1490,7 @@ double *EM_grad_lc(double *cc, int EM, int REML, int ns[])
     return cc;
 }
 
+#if 0
 
 /**
  * Print the verbose output in the ECME iterations
@@ -1500,8 +1501,9 @@ double *EM_grad_lc(double *cc, int EM, int REML, int ns[])
  * @param firstDer arrays for calculating ECME steps and the first derivative
  * @param val Pointer to a list of arrays to receive the calculated values
  */
-static
-void EMsteps_verbose_print(SEXP x, int iter, int REML, SEXP firstDer, SEXP val)
+static void
+EMsteps_verbose_print(SEXP x, int iter, int REML, SEXP firstDer,
+		      SEXP val)
 {
     SEXP Omega = GET_SLOT(x, Matrix_OmegaSym),
 	pMat = VECTOR_ELT(val, 2);
@@ -1554,44 +1556,83 @@ void EMsteps_verbose_print(SEXP x, int iter, int REML, SEXP firstDer, SEXP val)
     Rprintf("\n");
     Free(cc);
 }
+#endif
 
 /**
- * Perform ECME steps for the REML or ML criterion.
+ * Print the verbose output in the ECME iterations
  *
  * @param x pointer to an ssclme object
- * @param nsteps pointer to an integer scalar - the number of ECME steps to perform
- * @param Verbp pointer to a logical scalar indicating verbose output
- *
- * @return R_NilValue if verb == FALSE, otherwise a list of iteration
- *numbers, deviances, parameters, and gradients.
+ * @param iter iteration number
+ * @param REML non-zero for REML, zero for ML
+ * @param firstDer arrays for calculating ECME steps and the first derivative
+ * @param val Pointer to a list of arrays to receive the calculated values
  */
-SEXP lmer_ECMEsteps(SEXP x, SEXP nsteps, SEXP Verbp)
+static void
+EMsteps_verbose_print(SEXP x, int iter, int REML, SEXP firstDer)
+{
+    SEXP Omega = GET_SLOT(x, Matrix_OmegaSym);
+    int *nc = INTEGER(GET_SLOT(x, Matrix_ncSym)),
+	i, ifour = 4, ii, ione = 1, jj, nf = LENGTH(Omega);
+    double
+	*cc = EM_grad_lc(Calloc(4, double), 0, REML, nc + nf),
+	*dev = REAL(GET_SLOT(x, Matrix_devianceSym)),
+	one = 1., zero = 0.;
+
+    lmer_factor(x);
+    if (iter == 0) Rprintf("  EM iterations\n");
+    Rprintf("%3d %.3f", iter, dev[REML ? 1 : 0]);
+    for (i = 0; i < nf; i++) {
+	int nci = nc[i], ncip1 = nci + 1, ncisqr = nci * nci;
+	double
+	    *Omgi = REAL(VECTOR_ELT(Omega, i)),
+	    *Grad = Calloc(ncisqr, double);
+
+				/* diagonals */
+	Rprintf(" (%#8g", Omgi[0]);
+	for (jj = 1; jj < nci; jj++) {
+	    Rprintf(" %#8g", Omgi[jj * ncip1]);
+	}
+	for (jj = 1; jj < nci; jj++) /* offdiagonals */
+	    for (ii = 0; ii < jj; ii++)
+		Rprintf(" %#8g", Omgi[ii + jj * nci]);
+				/* Evaluate and print the gradient */
+	F77_CALL(dgemv)("N", &ncisqr, &ifour, &one,
+			REAL(VECTOR_ELT(firstDer, i)), &ncisqr,
+			cc, &ione, &zero, Grad, &ione);
+	Rprintf(":%#8.3g", Grad[0]);
+				
+	for (jj = 1; jj < nci; jj++) { /* diagonals */
+	    Rprintf(" %#8.3g", Grad[jj * ncip1]);
+	}
+	for (jj = 1; jj < nci; jj++) /* offdiagonals */
+	    for (ii = 0; ii < jj; ii++)
+		Rprintf(" %#8.3g", Grad[ii + jj * nci]);
+	Rprintf(")");
+	Free(Grad);
+    }
+    Rprintf("\n");
+    Free(cc);
+}
+
+static void
+internal_ECMEsteps(SEXP x, int nEM, int verb)
 {
     SEXP Omega = GET_SLOT(x, Matrix_OmegaSym),
-	flist = GET_SLOT(x, Matrix_flistSym),
-	val = R_NilValue;
+	flist = GET_SLOT(x, Matrix_flistSym);
     int *nc = INTEGER(GET_SLOT(x, Matrix_ncSym)),
 	*status = LOGICAL(GET_SLOT(x, Matrix_statusSym)),
-	REML = !strcmp(CHAR(asChar(GET_SLOT(x, Matrix_methodSym))), "REML"),
+	REML = !strcmp(CHAR(asChar(GET_SLOT(x, Matrix_methodSym))),
+		       "REML"),
 	i, ifour = 4, info, ione = 1, iter,
-	nEM = asInteger(nsteps),
-	nf = length(Omega),
-	verb = asLogical(Verbp);
+	nf = LENGTH(Omega);
     double
 	*cc = EM_grad_lc(Calloc(4, double), 1, REML, nc + nf),
 	zero = 0.0;
     SEXP firstDer = PROTECT(EM_grad_array(nf, nc));
 
     lmer_firstDer(x, firstDer);
-    if (verb) {
-	int nEMp1 = nEM + 1, npar = coef_length(nf, nc);
-	val = PROTECT(allocVector(VECSXP, 4));
-	SET_VECTOR_ELT(val, 0, allocVector(INTSXP, nEMp1));
-	SET_VECTOR_ELT(val, 1, allocVector(REALSXP, nEMp1));
-	SET_VECTOR_ELT(val, 2, allocMatrix(REALSXP, nEMp1, npar));
-	SET_VECTOR_ELT(val, 3, allocMatrix(REALSXP, nEMp1, npar));
-	EMsteps_verbose_print(x, 0, REML, firstDer, val);
-    }
+    if (verb)
+	EMsteps_verbose_print(x, 0, REML, firstDer);
     for (iter = 0; iter < nEM; iter++) {
 	for (i = 0; i < nf; i++) {
 	    int nci = nc[i], ncisqr = nci * nci;
@@ -1605,19 +1646,36 @@ SEXP lmer_ECMEsteps(SEXP x, SEXP nsteps, SEXP Verbp)
 			    cc, &ifour, &zero, Omgi, &ncisqr);
 	    F77_CALL(dpotrf)("U", &nci, Omgi, &nci, &info);
 	    if (info)
-		error(_("DPOTRF in ECME update gave code %d"), info);
+		error(_("DPOTRF in ECME update gave code %d"),
+		      info);
 	    F77_CALL(dpotri)("U", &nci, Omgi, &nci, &info);
 	    if (info)
 		error(_("Matrix inverse in ECME update gave code %d"), info);
 	}
 	status[0] = status[1] = 0;
 	lmer_firstDer(x, firstDer);
-	if (verb) EMsteps_verbose_print(x, iter + 1, REML, firstDer, val);
+	if (verb)
+	    EMsteps_verbose_print(x, iter + 1, REML, firstDer);
     }
     lmer_factor(x);
-    if (verb) UNPROTECT(1);
+    Free(cc);
     UNPROTECT(1);
-    return val;
+}
+
+/**
+ * Perform ECME steps for the REML or ML criterion.
+ *
+ * @param x pointer to an mer object
+ * @param nsteps pointer to an integer scalar - the number of ECME steps to perform
+ * @param Verbp pointer to a logical scalar indicating verbose output
+ *
+ * @return R_NilValue if verb == FALSE, otherwise a list of iteration
+ *numbers, deviances, parameters, and gradients.
+ */
+SEXP lmer_ECMEsteps(SEXP x, SEXP nsteps, SEXP Verbp)
+{
+    internal_ECMEsteps(x, asInteger(nsteps), asLogical(Verbp));
+    return R_NilValue;
 }
 
 /** 
@@ -1817,7 +1875,7 @@ SEXP lmer_secondDer(SEXP x, SEXP Valp)
  * @return a, symmetrized
  */
 static double*
-internal_symmetrize(double *a, const int nc)
+internal_symmetrize(double *a, int nc)
 {
     int i, j;
 
@@ -1858,6 +1916,34 @@ SEXP lmer_variances(SEXP x)
     return Omg;
 }
 
+static double*
+fitted_ranef(SEXP flist, SEXP mmats, SEXP b,
+	     const int nc[], double val[])
+{
+    int i, ione = 1, nf = LENGTH(flist),
+	nobs = LENGTH(VECTOR_ELT(flist, 0));
+    double one = 1;
+    
+    for (i = 0; i < nf; i++) {
+	SEXP bi = VECTOR_ELT(b, i);
+	int mi = INTEGER(getAttrib(bi, R_DimSymbol))[0];
+	int *ff = INTEGER(VECTOR_ELT(flist, i)), j, nci = nc[i];
+	double *mm = REAL(VECTOR_ELT(mmats, i));
+
+	for (j = 0; j < nobs; ) {
+	    int nn = 1, lev = ff[j];
+	    /* check for adjacent rows with same factor level */
+	    while ((j + nn) < nobs && ff[j + nn] == lev) nn++; 
+	    F77_CALL(dgemm)("N", "T", &nn, &ione, &nci,
+			    &one, mm + j, &nobs,
+			    REAL(bi) + (lev - 1), &mi,
+			    &one, &val[j], &nobs);
+	    j += nn;
+	}
+    }
+    return val;
+}
+    
 /** 
  * Calculate the fitted values.
  * 
@@ -1884,25 +1970,8 @@ internal_fitted(SEXP x, SEXP mmats, int useRand, double val[])
 			   REAL(fixef), &p, &zero, val, &nobs);
     else AZERO(val, nobs);
     if (useRand) {
-	int i;
-	SEXP b = PROTECT(lmer_ranef(x));
-	for (i = 0; i < nf; i++) {
-	    SEXP bi = VECTOR_ELT(b, i);
-	    int mi = INTEGER(getAttrib(bi, R_DimSymbol))[0];
-	    int *ff = INTEGER(VECTOR_ELT(flist, i)), j, nci = nc[i];
-	    double *mm = REAL(VECTOR_ELT(mmats, i));
-
-	    for (j = 0; j < nobs; ) {
-		int nn = 1, lev = ff[j];
-		/* check for adjacent rows with same factor level */
-		while ((j + nn) < nobs && ff[j + nn] == lev) nn++; 
-		F77_CALL(dgemm)("N", "T", &nn, &ione, &nci,
-				&one, mm + j, &nobs,
-				REAL(bi) + (lev - 1), &mi,
-				&one, &val[j], &nobs);
-		j += nn;
-	    }
-	}
+	fitted_ranef(flist, mmats, PROTECT(lmer_ranef(x)),
+		     nc, val);
 	UNPROTECT(1);
     }
     UNPROTECT(1);
@@ -1910,7 +1979,7 @@ internal_fitted(SEXP x, SEXP mmats, int useRand, double val[])
 }
 
 /** 
- * Return the fitted values.
+ * Return the fitted values as an SEXP
  * 
  * @param x pointer to an lmer object
  * @param mmats list of model matrices
@@ -1930,54 +1999,7 @@ SEXP lmer_fitted(SEXP x, SEXP mmats, SEXP useRf)
     return ans;
 }
 
-/** 
- * Copy an lmer object collapsing the fixed effects slots to the response only.
- * 
- * @param x pointer to an lmer object
- * 
- * @return a duplicate of x with the fixed effects slots collapsed to the response only
- */
-SEXP lmer_collapse(SEXP x)
-{
-    SEXP 
-        ans = PROTECT(NEW_OBJECT(MAKE_CLASS("lmer"))),
-	Omega = GET_SLOT(x, Matrix_OmegaSym),
-	Dim = getAttrib(GET_SLOT(x, Matrix_ZtXSym), R_DimSymbol);
-    int 
-        nf = length(Omega), 
-        nz = INTEGER(Dim)[0];
-
-    slot_dup(ans, x, Matrix_flistSym);
-    slot_dup(ans, x, Matrix_permSym);
-    slot_dup(ans, x, Matrix_ParentSym);
-    slot_dup(ans, x, Matrix_DSym);
-    slot_dup(ans, x, Matrix_bVarSym);
-    slot_dup(ans, x, Matrix_LSym);
-    slot_dup(ans, x, Matrix_ZZpOSym);
-    slot_dup(ans, x, Matrix_OmegaSym);
-    slot_dup(ans, x, Matrix_methodSym);
-    slot_dup(ans, x, Matrix_ZtZSym);
-
-    slot_dup(ans, x, Matrix_cnamesSym);
-    slot_dup(ans, x, Matrix_devCompSym);
-    slot_dup(ans, x, Matrix_devianceSym);
-    slot_dup(ans, x, Matrix_ncSym);
-    slot_dup(ans, x, Matrix_GpSym);
-    slot_dup(ans, x, Matrix_statusSym);
-
-    INTEGER(GET_SLOT(ans, Matrix_ncSym))[nf] = 1;
-    SET_SLOT(ans, Matrix_XtXSym, allocMatrix(REALSXP, 1, 1));
-    REAL(GET_SLOT(ans, Matrix_XtXSym))[0] = NA_REAL;
-    SET_SLOT(ans, Matrix_RXXSym, allocMatrix(REALSXP, 1, 1));
-    REAL(GET_SLOT(ans, Matrix_RXXSym))[0] = NA_REAL;
-    SET_SLOT(ans, Matrix_ZtXSym, allocMatrix(REALSXP, nz, 1));
-    SET_SLOT(ans, Matrix_RZXSym, allocMatrix(REALSXP, nz, 1));
-    LOGICAL(GET_SLOT(ans, Matrix_statusSym))[0] = 0;
-    UNPROTECT(1);
-    return ans;
-}
-
-/* Gauss-Hermite Quadrature x positions */
+/* Gauss-Hermite Quadrature x positions and weights */
 static const double
     GHQ_x1[1] = {0},
     GHQ_w1[1] = {1},
@@ -1988,35 +2010,47 @@ static const double
     GHQ_x4[2] = {2.3344141783872, 0.74196377160456},
     GHQ_w4[2] = {0.0458758533899086, 0.454124131589555},
     GHQ_x5[3] = {2.85696996497785, 1.35562615677371, 0},
-    GHQ_w5[3] = {0.0112574109895360, 0.222075915334214, 0.533333317311434},
-    GHQ_x6[3] = {3.32425737665988, 1.88917584542184, 0.61670657963811},
-    GHQ_w6[3] = {0.00255578432527774, 0.0886157433798025, 0.408828457274383},
-    GHQ_x7[4] = {3.7504396535397, 2.36675937022918, 1.15440537498316, 0},
-    GHQ_w7[4] = {0.000548268839501628, 0.0307571230436095, 0.240123171391455,
-		 0.457142843409801},
-    GHQ_x8[4] = {4.14454711519499, 2.80248581332504, 1.63651901442728,
-		 0.539079802125417},
-    GHQ_w8[4] = {0.000112614534992306, 0.0096352198313359, 0.117239904139746,
-		 0.373012246473389},
-    GHQ_x9[5] = {4.51274578616743, 3.20542894799789, 2.07684794313409,
-		 1.02325564627686, 0},
-    GHQ_w9[5] = {2.23458433364535e-05, 0.00278914123744297, 0.0499164052656755,
-		 0.244097495561989, 0.406349194142045},
-    GHQ_x10[5] = {4.85946274516615, 3.58182342225163, 2.48432579912153,
-		  1.46598906930182, 0.484935699216176},
-    GHQ_w10[5] = {4.31065250122166e-06, 0.000758070911538954, 0.0191115799266379,
-		  0.135483698910192, 0.344642324578594},
-    GHQ_x11[6] = {5.18800113558601, 3.93616653976536, 2.86512311160915,
-		  1.87603498804787, 0.928868981484148, 0},
-    GHQ_w11[6] = {8.12184954622583e-07, 0.000195671924393029, 0.0067202850336527,
-		  0.066138744084179, 0.242240292596812, 0.36940835831095};
+    GHQ_w5[3] = {0.0112574109895360, 0.222075915334214,
+		 0.533333317311434},
+    GHQ_x6[3] = {3.32425737665988, 1.88917584542184,
+		 0.61670657963811},
+    GHQ_w6[3] = {0.00255578432527774, 0.0886157433798025,
+		 0.408828457274383},
+    GHQ_x7[4] = {3.7504396535397, 2.36675937022918,
+		 1.15440537498316, 0},
+    GHQ_w7[4] = {0.000548268839501628, 0.0307571230436095,
+		 0.240123171391455, 0.457142843409801},
+    GHQ_x8[4] = {4.14454711519499, 2.80248581332504,
+		 1.63651901442728, 0.539079802125417},
+    GHQ_w8[4] = {0.000112614534992306, 0.0096352198313359,
+		 0.117239904139746, 0.373012246473389},
+    GHQ_x9[5] = {4.51274578616743, 3.20542894799789,
+		 2.07684794313409, 1.02325564627686, 0},
+    GHQ_w9[5] = {2.23458433364535e-05, 0.00278914123744297,
+		 0.0499164052656755, 0.244097495561989,
+		 0.406349194142045},
+    GHQ_x10[5] = {4.85946274516615, 3.58182342225163,
+		  2.48432579912153, 1.46598906930182,
+		  0.484935699216176},
+    GHQ_w10[5] = {4.31065250122166e-06, 0.000758070911538954,
+		  0.0191115799266379, 0.135483698910192,
+		  0.344642324578594},
+    GHQ_x11[6] = {5.18800113558601, 3.93616653976536,
+		  2.86512311160915, 1.87603498804787,
+		  0.928868981484148, 0},
+    GHQ_w11[6] = {8.12184954622583e-07, 0.000195671924393029,
+		  0.0067202850336527, 0.066138744084179,
+		  0.242240292596812, 0.36940835831095};
 
 static const double
-    *GHQ_x[12] = {(double *) NULL, GHQ_x1, GHQ_x2, GHQ_x3, GHQ_x4, GHQ_x5,
-		  GHQ_x6, GHQ_x7, GHQ_x8, GHQ_x9, GHQ_x10, GHQ_x11},
-    *GHQ_w[12] = {(double *) NULL, GHQ_w1, GHQ_w2, GHQ_w3, GHQ_w4, GHQ_w5,
-		  GHQ_w6, GHQ_w7, GHQ_w8, GHQ_w9, GHQ_w10, GHQ_w11};
+    *GHQ_x[12] = {(double *) NULL, GHQ_x1, GHQ_x2, GHQ_x3, GHQ_x4,
+		  GHQ_x5, GHQ_x6, GHQ_x7, GHQ_x8, GHQ_x9, GHQ_x10,
+		  GHQ_x11},
+    *GHQ_w[12] = {(double *) NULL, GHQ_w1, GHQ_w2, GHQ_w3, GHQ_w4,
+		  GHQ_w5, GHQ_w6, GHQ_w7, GHQ_w8, GHQ_w9, GHQ_w10,
+		  GHQ_w11};
 
+#if 0
 /** 
  * Compute certain components of the Laplace likelihood approximation 
  * 
@@ -2072,9 +2106,12 @@ SEXP glmer_Laplace_devComp(SEXP x) {
     UNPROTECT(1);
     return ScalarReal(ans);
 }
+
+#endif
 				 
 static void
-glmer_wt_lst(SEXP MLin, double *wts, double *adjst, int n, SEXP MLout)
+internal_weight_list(SEXP MLin, double *wts, double *adjst, int n,
+		     SEXP MLout)
 {
     int i, j, nf = LENGTH(MLin);
     SEXP lastM;
@@ -2120,7 +2157,8 @@ SEXP find_and_check(SEXP rho, SEXP nm, SEXPTYPE mode, int len)
 	error(_("environment `rho' must contain an object `%s'"),
 	      CHAR(PRINTNAME(nm)));
     if (TYPEOF(ans) != mode)
-	error(_("object `%s' of incorrect type"), CHAR(PRINTNAME(nm)));
+	error(_("object `%s' of incorrect type"),
+	      CHAR(PRINTNAME(nm)));
     if (len && LENGTH(ans) != len)
 	error(_("object `%s' must be of length `%d'"),
 	      CHAR(PRINTNAME(nm)), len);
@@ -2152,9 +2190,8 @@ SEXP eval_check_store(SEXP fcn, SEXP rho, SEXP vv)
     return vv;
 }
 
-static
-SEXP eval_check(SEXP fcn, SEXP rho, SEXPTYPE mode, int len)
-{
+static SEXP
+eval_check(SEXP fcn, SEXP rho, SEXPTYPE mode, int len) {
     SEXP v = PROTECT(eval(fcn, rho));
     if (TYPEOF(v) != mode || LENGTH(v) != len)
 	error(_("fcn produced mode %d, length %d - wanted mode %d, length %d"),
@@ -2163,119 +2200,324 @@ SEXP eval_check(SEXP fcn, SEXP rho, SEXPTYPE mode, int len)
     return v;
 }
 
-#define BHAT_NITER 20
+static SEXP
+getElement(SEXP list, char *nm) {
+    SEXP names = getAttrib(list, R_NamesSymbol);
+    int i;
 
-SEXP lmer_devLaplace(SEXP pars, SEXP tolp, SEXP rho)
+    for (i = 0; i < LENGTH(list); i++)
+	if (!strcmp(CHAR(STRING_ELT(names, i)), nm))
+	    return(VECTOR_ELT(list, i));
+    return R_NilValue;
+}
+
+typedef struct glmer_struct
 {
-    SEXP dev_resids, eta, linkinv, mu, mu_eta, offset, rdobj,
-	unwtd, variance, wts, wtd, x, y;
-    int conv, i, ii, ione = 1, n, *nc, nf, p;
-    SEXP dmu_deta, var, devr;
-    double *etaold, *fitted, *off, *w, *z, ans,
-	one = 1, tol = asReal(tolp), zero = 0;
-	
-    if (!isReal(pars) || LENGTH(pars) < 1)
-	error(_("`%s' must be a nonempty, numeric vector"), "pars");
+    SEXP cv;         /* control values */
+    SEXP mer;	     /* mixed-effects representation */
+    SEXP rho;        /* where to evaluate the calls */
+    SEXP y;          /* response vector (REALSXP) */
+    SEXP eta;        /* linear predictor */
+    SEXP mu;         /* mean vector */
+    SEXP wts;        /* prior weights */
+    SEXP offset;     /* offset for original glm problem */
+    SEXP off;        /* offset for random effects only */
+    SEXP unwtd;      /* unweighted model matrices */
+    SEXP wtd;        /* weighted model matrices */
+    SEXP linkinv;    /* expression for inverse link evaluation */
+    SEXP mu_eta;     /* expression for dmu/deta evaluation */
+    SEXP LMEopt;     /* expression for LME optimization */
+    SEXP dev_resids; /* expression for deviance residuals */
+    SEXP var;        /* expression for variance evaluation */
+    SEXP x;          /* fixed-effects model matrix */
+    int n;	     /* length of the response vector */
+    int p;	     /* length of fixed effects vector */
+    int nf;	     /* number of grouping factors */
+    int npar;        /* total length of the parameter */
+    int niterEM;     /* default number of ECME iterations */
+    int EMverbose;   /* logical indicator */
+    int maxiter;     /* maximum number of IRLS iterations */
+    double tol;      /* convergence tolerance for IRLS iterations */
+} glmer_struct, *GlmerStruct;
+
+SEXP glmer_init(SEXP rho) {
+    GlmerStruct GS;
+    int *dims;
+    SEXP wtdSym = install("wtd"), offSym = install("off");
+    
+    GS = (GlmerStruct) Calloc(1, glmer_struct);
     if (!isEnvironment(rho))
 	error(_("`rho' must be an environment"));
-    y = find_and_check(rho, install("y"), REALSXP, 0);
-    if ((n = LENGTH(y)) < 1)
-	error(_("`%s' must be a nonempty, numeric vector"), "y");	
-    etaold = Calloc(n, double);
-    fitted = Calloc(n, double);
-    off = Calloc(n, double);
-    w = Calloc(n, double);
-    z = Calloc(n, double);
-    rdobj = find_and_check(rho, install("mer"), VECSXP, 0);
-    nc = INTEGER(GET_SLOT(rdobj, Matrix_ncSym));
-    nf = LENGTH(GET_SLOT(rdobj, Matrix_flistSym));
-    p = LENGTH(pars) - coef_length(nf, nc);
-    mu = find_and_check(rho, install("mu"), REALSXP, n);
-    offset = find_and_check(rho, install("offset"), REALSXP, n);
-    x = find_and_check(rho, install("x"), REALSXP, n * p);
-    eta = find_and_check(rho, install("eta"), REALSXP, n);
-    unwtd = find_and_check(rho, install("mmo"), VECSXP, nf + 1);
-    wts = find_and_check(rho, install("wts"), REALSXP, n);
-    wtd = find_and_check(rho, install("mmats"), VECSXP, nf + 1);
+    GS->rho = rho;
+    GS->y = find_and_check(rho, install("y"), REALSXP, 0);
+    if ((GS->n = LENGTH(GS->y)) < 1)
+	error(_("`%s' must be a nonempty, numeric vector"),
+	      "y");
+    GS->mer = find_and_check(rho, install("mer"), VECSXP, 0);
+    GS->nf = LENGTH(GET_SLOT(GS->mer, Matrix_flistSym));
+    GS->cv = find_and_check(rho, install("cv"), VECSXP, 0);
+    GS->niterEM = asInteger(getElement(GS->cv, "niterEM"));
+    GS->EMverbose = asLogical(getElement(GS->cv, "EMverbose"));
+    GS->tol = asReal(getElement(GS->cv, "tolerance"));
+    GS->maxiter = asInteger(getElement(GS->cv, "maxIter"));
 
-    dev_resids = find_and_check(rho, install("dev.resids"), LANGSXP, 0);
-    linkinv = find_and_check(rho, install("linkinv"), LANGSXP, 0);
-    mu_eta = find_and_check(rho, install("mu.eta"), LANGSXP, 0);
-    variance = find_and_check(rho, install("variance"), LANGSXP, 0);
+    GS->mu = find_and_check(rho, install("mu"), REALSXP, GS->n);
+    GS->offset = find_and_check(rho, install("offset"),
+				REALSXP, GS->n);
+    defineVar(offSym, duplicate(GS->offset), rho);
+    GS->off = find_and_check(rho, install("off"), REALSXP, GS->n);
+    GS->x = find_and_check(rho, install("x"), REALSXP, 0);
+    if (!isMatrix(GS->x))
+	error(_("%s must be a model matrix"), "x");
+    dims = INTEGER(getAttrib(GS->x, R_DimSymbol));
+    if (dims[0] != GS->n)
+	error(_("nrow(%s) must be %d"), "x", GS->n);
+    GS->p = dims[1];
+    GS->npar = dims[1] +
+	coef_length(GS->nf, INTEGER(GET_SLOT(GS->mer, Matrix_ncSym)));
+    GS->eta = find_and_check(rho, install("eta"),
+			     REALSXP, GS->n);
+    GS->unwtd = find_and_check(rho, install("mmats"),
+			       VECSXP, GS->nf + 1);
+    defineVar(wtdSym, duplicate(GS->unwtd), rho);
+    GS->wtd = find_and_check(rho, wtdSym, VECSXP, GS->nf + 1);
+    GS->wts = find_and_check(rho, install("wts"), REALSXP, GS->n);
+    GS->linkinv = find_and_check(rho, install("linkinv"),
+				 LANGSXP, 0);
+    GS->mu_eta = find_and_check(rho, install("mu.eta"),
+				LANGSXP, 0);
+    GS->var = find_and_check(rho, install("variance"),
+			     LANGSXP, 0);
+    GS->LMEopt = find_and_check(rho, install("doLMEopt"),
+				LANGSXP, 0);
+    GS->dev_resids = find_and_check(rho, install("dev.resids"),
+				    LANGSXP, 0);
+    return R_MakeExternalPtr(GS, R_NilValue, GS->mer);
+}
 
-    F77_CALL(dgemv)("N", &n, &p, &one, REAL(x), &n,
-		    REAL(pars), &ione, &zero, off, &ione);
-    for (ii = 0; ii < n; ii++)
-	REAL(eta)[ii] = (off[ii] += REAL(offset)[ii]);
-    internal_coefGets(rdobj, REAL(pars) + p, 2);
-    conv = 0;
+SEXP glmer_finalize(SEXP GSp) {
+    GlmerStruct GS = (GlmerStruct) R_ExternalPtrAddr(GSp);
     
-    Memcpy(etaold, REAL(eta), n);
-    for (i = 0; i < BHAT_NITER && !conv; i++) {
-	double max_eta, max_diff;
+    Free(GS);
+    return R_NilValue;
+}
 
-	eval_check_store(linkinv, rho, mu);
-	dmu_deta = PROTECT(eval_check(mu_eta, rho, REALSXP, n));
-	var = PROTECT(eval_check(variance, rho, REALSXP, n));
-	for (ii = 0; ii < n; ii++) {
-	    w[ii] = REAL(wts)[ii] * REAL(dmu_deta)[ii]/sqrt(REAL(var)[ii]);
-	    z[ii] = REAL(eta)[ii] - off[ii] +
-		(REAL(y)[ii] - REAL(mu)[ii])/REAL(dmu_deta)[ii];
-	}
-	UNPROTECT(2);
-	glmer_wt_lst(unwtd, w, z, n, wtd);
-	lmer_update_mm(rdobj, wtd);
-	internal_fitted(rdobj, unwtd, 1, fitted);
-	max_eta = max_diff = -1.;
-	for (ii = 0; ii < n; ii++) {
-	    double abs_eta, abs_diff;
+/** 
+ * Evaluate new weights and working residuals, apply the weights to a
+ * copy of the model matrices and update the mer object.
+ * 
+ * @param GS a GlmerStruct object
+ */
+static void
+reweight_update(GlmerStruct GS) {
+    SEXP dmu_deta, var;
+    int i;
+    double *w = Calloc(GS->n, double),
+	*z = Calloc(GS->n, double);
+				/* reweight mer */
+    eval_check_store(GS->linkinv, GS->rho, GS->mu);
+    dmu_deta = PROTECT(eval_check(GS->mu_eta, GS->rho,
+				  REALSXP, GS->n));
+    var = PROTECT(eval_check(GS->var, GS->rho,
+			     REALSXP, GS->n));
+    for (i = 0; i < GS->n; i++) {
+	w[i] = REAL(GS->wts)[i] *
+	    REAL(dmu_deta)[i]/sqrt(REAL(var)[i]);
+	z[i] = REAL(GS->eta)[i] - REAL(GS->off)[i] +
+	    (REAL(GS->y)[i] - REAL(GS->mu)[i]) /
+	    REAL(dmu_deta)[i];
+    }
+    internal_weight_list(GS->unwtd, w, z, GS->n, GS->wtd);
+    lmer_update_mm(GS->mer, GS->wtd);
+    UNPROTECT(2);
+    Free(w); Free(z);
+}
+
+/** 
+ * Update eta, evaluate the convergence criterion, then copy eta to
+ * etaold
+ * 
+ * @param GS a GlmerStruct object
+ * @param etaold previous values of the linear predictors
+ * 
+ * @return convergence criterion
+ */
+static double
+conv_crit(GlmerStruct GS, double etaold[]) {
+    double *fitted = Calloc(GS->n, double),
+	max_eta = -1, max_diff = -1;
+    int i;
+
+    internal_fitted(GS->mer, GS->unwtd, 1, fitted);
+    for (i = 0; i < GS->n; i++) {
+	double abs_eta, abs_diff;
 	    
-	    REAL(eta)[ii] = off[ii] + fitted[ii];
-	    abs_eta = fabs(REAL(eta)[ii]);
-	    if (abs_eta > max_eta) max_eta = abs_eta;
-	    abs_diff = fabs(REAL(eta)[ii] - etaold[ii]);
-	    if (abs_diff > max_diff) max_diff = abs_diff;
-	    etaold[ii] = REAL(eta)[ii];
-	}
-	if (max_diff < (0.1 + max_eta) * tol) {
-	    conv = 1;
-	    break;
-	}
+	REAL(GS->eta)[i] = REAL(GS->off)[i] + fitted[i];
+	abs_eta = fabs(REAL(GS->eta)[i]);
+	if (abs_eta > max_eta) max_eta = abs_eta;
+	abs_diff = fabs(REAL(GS->eta)[i] - etaold[i]);
+	if (abs_diff > max_diff) max_diff = abs_diff;
+	etaold[i] = REAL(GS->eta)[i];
+    }
+    Free(fitted);
+
+    return max_diff / (0.1 + max_eta);
+}
+
+/** 
+ * Perform the PQL optimization
+ * 
+ * @param GSp pointer to a GlmerStruct object
+ * 
+ * @return R_NilValue
+ */
+SEXP glmer_PQL(SEXP GSp)
+{
+    GlmerStruct GS = (GlmerStruct) R_ExternalPtrAddr(GSp);
+    int conv, i;
+    double *etaold = Memcpy(Calloc(GS->n, double),
+			    REAL(GS->eta), GS->n);
+
+    for (i = 0, conv = 0; i < GS->maxiter && !conv; i++) {
+	reweight_update(GS);
+	if (!i) lmer_initial(GS->mer); /* initialize first fit */
+	internal_ECMEsteps(GS->mer, i ? 2 : GS->niterEM,
+			   GS->EMverbose);
+	eval(GS->LMEopt, GS->rho);
+	conv = conv_crit(GS, etaold) < GS->tol;
+    }
+    if (!conv)
+	warning(_("IRLS iterations for PQL did not converge"));
+
+    Free(etaold);
+    return R_NilValue;
+}
+
+/** 
+ * Evaluate half the difference of the logarithm of the
+ * determinant of Sigma and the log determinant of bVar.
+ * This function has a side effect of factoring Sigma and the
+ * matrices in bVar.
+ * 
+ * @param GS a GlmerStruct object
+ * @param b a random-effects object
+ * 
+ * @return difference of log determinants
+ */
+static double
+Sigma_bVar_det(GlmerStruct GS) {
+    SEXP Omega = GET_SLOT(GS->mer, Matrix_OmegaSym),
+	bVar = GET_SLOT(GS->mer, Matrix_bVarSym);
+    int *nc = INTEGER(GET_SLOT(GS->mer, Matrix_ncSym)),
+	*Gp = INTEGER(GET_SLOT(GS->mer, Matrix_GpSym)), i;
+    double ans;
+
+    for (i = 0, ans = 0; i < GS->nf; i++) {
+        int j, k, nci = nc[i];
+        int ncip1 = nci + 1, ncisqr = nci * nci,
+	    nlev = (Gp[i + 1] - Gp[i])/nci;
+	double *bVi = REAL(VECTOR_ELT(bVar, i)),
+	    *Omgi = REAL(VECTOR_ELT(Omega, i));
+
+        F77_CALL(dpotrf)("U", &nci, Omgi, &nci, &j);
+        if (j)
+            error(_("Leading %d minor of Omega[[%d]] not positive definite"),
+                  j, i + 1);
+        for (j = 0; j < nci; j++) { /* nlev * logDet(Omega_i) */
+            ans += nlev * log(Omgi[j * ncip1]);
+            /* Note: we are adding because Omega has been inverted */
+        }
+        for (k = 0; k < nlev; k++) {
+	    double *bVik = bVi + k * ncisqr;
+            F77_CALL(dpotrf)("U", &nci, bVik, &nci, &j);
+            if (j)
+                error(_("Leading %d minor of bVar[[%d]][,,%d] not positive definite"),
+                      j, i + 1, k + 1);
+            for (j = 0; j < nci; j++) ans += log(bVik[j * ncip1]);
+        }
+    }
+    return ans;
+}
+
+/** 
+ * Establish off, the effective offset for the fixed effects, and
+ * iterate to determine the conditional modes.  Factor Omega and bVar
+ * and return the difference in the log-determinants.
+ * 
+ * @param pars parameter vector
+ * @param GS a GlmerStruct object
+ */
+static void
+glmer_bhat(double pars[], GlmerStruct GS)
+{
+    int conv, i, ione = 1;
+    double *etaold, one = 1, zero = 0;
+	
+    F77_CALL(dgemv)("N", &(GS->n), &(GS->p), &one,
+		    REAL(GS->x), &(GS->n), pars,
+		    &ione, &zero, REAL(GS->off), &ione);
+    for (i = 0; i < GS->n; i++)
+	REAL(GS->eta)[i] =
+	    (REAL(GS->off)[i] += REAL(GS->offset)[i]);
+    internal_coefGets(GS->mer, pars + GS->p, 2);
+    etaold = Memcpy(Calloc(GS->n, double), REAL(GS->eta), GS->n);
+
+    for (i = 0, conv = 0; i < GS->maxiter && !conv; i++) {
+	reweight_update(GS);
+	conv = conv_crit(GS, etaold) < GS->tol;
     }
     if (!conv) warning(_("iterations for bhat did not converge"));
-    Free(etaold); Free(fitted); Free(off); Free(w); Free(z);
-    devr = PROTECT(eval_check(dev_resids, rho, REALSXP, n));
-    ans = 0;
-    for (ii = 0; ii < n; ii++) ans += REAL(devr)[ii];
-    UNPROTECT(1);
-    return ScalarReal(ans - 2 * asReal(glmer_Laplace_devComp(rdobj)));
+    Free(etaold);
 }
 
-/**
- * Produce a weighted copy of the matrices in MLin in the storage
- * allocated to MLout
- *
- * @param MLin input matrix list
- * @param wts real vector of weights
- * @param adjst adjusted response
- * @param MLout On input a list of matrices of the same dimensions as MLin.
- *
- * @return MLout with its contents overwritten by a weighted copy of
- * MLin according to wts with adjst overwriting the response.
- */
-SEXP glmer_weight_matrix_list(SEXP MLin, SEXP wts, SEXP adjst, SEXP MLout)
-{
-    int n, nf;
+static double
+b_quadratic_form(GlmerStruct GS, SEXP b) {
+    SEXP Omega = GET_SLOT(GS->mer, Matrix_OmegaSym);
+    int *nc = INTEGER(GET_SLOT(GS->mer, Matrix_ncSym)),
+	*Gp = INTEGER(GET_SLOT(GS->mer, Matrix_GpSym)),
+	i, ione = 1;
+    double ans, one = 1, zero = 0;
     
-    if (!(isNewList(MLin) && isReal(wts) && isReal(adjst) && isNewList(MLout)))
-	error(_("Incorrect argument type"));
-    nf = LENGTH(MLin);
-    if (LENGTH(MLout) != nf)
-	error(_("Lengths of MLin (%d) and MLout (%d) must match"), nf,
-	      LENGTH(MLout));
-    n = LENGTH(wts);
-    if (LENGTH(adjst) != n)
-	error(_("Expected adjst to have length %d, got %d"), n, LENGTH(adjst));
-    glmer_wt_lst(MLin, REAL(wts), REAL(adjst), n, MLout);
-    return MLout;
+    for (i = 0, ans = 0; i < GS->nf; i++) {
+	int nci = nc[i], ntot = Gp[i + 1] - Gp[i];
+	int nlev = ntot/nci;
+	double *tmp = Calloc(ntot, double);
+
+	F77_CALL(dgemm)("N", "T", &nlev, &nci, &nci, &one,
+			REAL(VECTOR_ELT(b, i)), &nlev,
+			REAL(VECTOR_ELT(Omega, i)), &nci,
+			&zero, tmp, &nlev);
+        ans += F77_CALL(ddot)(&ntot, tmp, &ione, tmp, &ione);
+    }
+    return ans;
 }
+	
+static double
+marginal_deviance(GlmerStruct GS) {
+    SEXP devr = PROTECT(eval_check(GS->dev_resids, GS->rho,
+				   REALSXP, GS->n));
+    int i;
+    double ans = 0;
+    
+    for (i = 0; i < GS->n; i++) ans += REAL(devr)[i];
+    UNPROTECT(1);
+    return ans;
+}
+
+SEXP glmer_devAGQ(SEXP pars, SEXP GSp, SEXP nAGQp)
+{
+    GlmerStruct GS = (GlmerStruct) R_ExternalPtrAddr(GSp);
+    int nAGQ = asInteger(nAGQp);
+	
+    if (!isReal(pars) || LENGTH(pars) != GS->npar)
+	error(_("`%s' must be a numeric vector of length %d"),
+	      "pars", GS->npar);
+    glmer_bhat(REAL(pars), GS);
+    if (nAGQ > 1) error("code not yet written");
+
+    return ScalarReal(marginal_deviance(GS)
+		      - 2 * Sigma_bVar_det(GS)
+		      - b_quadratic_form(GS, lmer_ranef(GS->mer)));
+}
+
+
